// Copyright (c) 2006 Advanced Micro Devices ("AMD").
//
// All Rights Reserved.  Unpublished rights reserved under the 
// copyright laws of the United States of America, other countries 
// and international treaties.  The software is provided without 
// fee.  Permission to use, copy, store, modify, disclose, transmit 
// or distribute the software is granted, provided that this 
// copyright notice must appear in any copy, modification, 
// disclosure, transmission or distribution of the software.
//
// AMD retains all ownership, copyright, trade secret and 
// proprietary rights in the software. THIS SOFTWARE HAS BEEN 
// PROVIDED "AS IS," WITHOUT EXPRESS OR IMPLIED WARRANTY INCLUDING, 
// WITHOUT LIMITATION, IMPLIED WARRANTIES OF MERCHANTABILITY, 
// FITNESS FOR A PARTICULAR USE AND NON-INFRINGEMENT.
//
// AMD does not assume or authorize any other person to assume for 
// it any liability in connection with the Software. AMD SHALL NOT 
// BE LIABLE TO COMPANY, OR ANY THIRD PARTY, IN CONTRACT, TORT, 
// WARRANTY, STRICT LIABILITY, OR OTHERWISE FOR ANY DIRECT DAMAGES, 
// OR FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR CONSEQUENTIAL 
// DAMAGES, INCLUDING BUT NOT LIMITED TO, BUSINESS INTERRUPTION, 
// LOST PROFITS OR GOODWILL, OR LOSS OF INFORMATION EVEN IF AMD IS 
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

#include <windows.h>

// GeodeLX Display Registry structure
// This structure holds all user-modifiable registry
// information, such as panel enable, custom timings, etc.  This
// structure definition is taken from lxshared.h in the display
// driver source.  It is duplicated here to allow this code to be
// compilable and functional standalone.
//
typedef struct tagLXRegistrySettings
{
    unsigned long secretWord;           // Identifier to indicate valid buffer
    unsigned long unused;               // Unused DWORD
    unsigned long BIOSDetectOutput;     // Use the BIOS to detect the output device.
    unsigned long Compression;          // Compression Settings (enabled, auto, etc.)
    unsigned long FBLinear;             // Linear frame buffer setting
    unsigned long DDCEnabled;           // Enable DDC communication to a monitor
    unsigned long DDC_GPIOData;         // GPIO for DDC data line.
    unsigned long DDC_GPIOClock;        // GPIO for DDC clock line.
    unsigned long PanelEnable;          // Dynamic state of panel enable.
    unsigned long PanelWidth;           // Manually specified panel width for supported panel sizes.
    unsigned long PanelHeight;          // Manually specified panel height for supported panel sizes.
    unsigned long XVGAPanel;            // XVGA Panel for supported panel sizes.
    unsigned long SimultaneousCRT;      // Dynamic flag enabling simultaneous CRT display.    
    unsigned long ManualTimingEnable;   // Custom mode timings have been specified.    
    unsigned long EnablePanelScale;     // Allow modes to be stretched to fit the display.
    unsigned long EnableNon60HzPanel;   // Allow non-60Hz CRT modes when displaying on a TFT.
    unsigned long CustomPanelTimings;   // Custom panel timings have been specified.
    unsigned long RotationAngle;        // Angle of desktop rotation.
    unsigned long PanelRefresh;         // Detected refresh rate of the panel
    unsigned long PanelInSystem;        // Panel detected with BIOSDetectOutput or panel
                                        // specifically specified at install time.

} LX_REGISTRY_SETTINGS;

// Driver Escape Definitions
// These definitions are also pulled from lxshared.h in the display driver
// source.
//
#define GEODELX_GET_REGISTRY_SETTINGS  0x400
#define GEODELX_SET_REGISTRY_SETTINGS  0x401
#define GEODELX_SECRET_WORD            0xABADD00D

// Global Variables
//
LX_REGISTRY_SETTINGS g_regData, g_oldSettings;
DEVMODE g_curMode;
DEVMODE g_newMode;

/////////////////////////////////////////////////////////////////////////////
// Rotate
//
// Routine to rotate the desktop to 0, 90, 180 or 270 degrees.  This routine,
// in addition to making the appropriate registry updates, will set the
// appropriate display mode and try to open the display settings dialog, if
// it is open.
//
int Rotate (int angle)
{
    BOOL retval;
    int mode;
    DEVMODE dm;
    LONG modeset;
    HDC dc;

    // Check for valid rotation angle.
    //
    if (angle != 0 && angle != 90 && angle != 180 && angle != 270)
        return -1;

    // Before we do anything, save a copy of the existing driver registry 
    // settings, such that we can restore them if the modeset fails.
    //
    dc = GetDC(NULL);
    g_oldSettings.secretWord = GEODELX_SECRET_WORD;
    retval = ExtEscape (dc, GEODELX_GET_REGISTRY_SETTINGS, sizeof(g_oldSettings), 
        (LPCSTR)&g_oldSettings, sizeof(g_oldSettings), (LPSTR)&g_oldSettings);
    ReleaseDC (NULL, dc);
    g_regData = g_oldSettings;

    // Verify valid communication with the driver.
    //
    if (retval <= 0)
        return -2;

    // Read the current display mode.  We need this information to figure out
    // the new mode that should be set.
    //
    EnumDisplaySettings (NULL, ENUM_CURRENT_SETTINGS, &g_curMode);

    // Keep the original around in case of a failed modeset, as we will
    // want to put the original mode back into the registry.  Copy the
    // contents into a new buffer that will contain the updated mode.
    //
    g_newMode = g_curMode;
    if (g_oldSettings.RotationAngle == 90 ||
        g_oldSettings.RotationAngle == 270)
    {
        unsigned long temp = g_newMode.dmPelsWidth;
        g_newMode.dmPelsWidth = g_newMode.dmPelsHeight;
        g_newMode.dmPelsHeight = temp;
    }

    // Update the registry settings.  This will potentially alter the list
    // of available modes.
    //
    dc = GetDC(NULL);
    g_regData.secretWord = GEODELX_SECRET_WORD;
    g_regData.RotationAngle = angle;
    retval = ExtEscape (dc, GEODELX_SET_REGISTRY_SETTINGS, sizeof(g_regData), 
        (LPCSTR)&g_regData, sizeof(g_regData), (LPSTR)&g_regData);
    ReleaseDC (NULL, dc);
        
    // Adjust the display mode for the new rotation angle.
    //
    if (g_regData.RotationAngle == 90 ||
        g_regData.RotationAngle == 270)
    {
        long temp = g_newMode.dmPelsWidth;
        g_newMode.dmPelsWidth = g_newMode.dmPelsHeight;
        g_newMode.dmPelsHeight = temp;
    }
    switch (g_regData.RotationAngle)
    {
        case 90:  g_newMode.dmDisplayOrientation = DMDO_90;      break;
        case 180: g_newMode.dmDisplayOrientation = DMDO_180;     break;
        case 270: g_newMode.dmDisplayOrientation = DMDO_270;     break;
        default:  g_newMode.dmDisplayOrientation = DMDO_DEFAULT; break;
    }

    // Enumerate all display modes.  This forces calls to DrvGetModes, which allows the call 
    // to set the rotated mode to work.
    //
    mode = 0;
    while (EnumDisplaySettings (NULL, mode, &dm))
        mode++;

    // Change the hardware display mode.  A modeset is needed to actually make the
    // rotation take effect.
    //
    modeset = ChangeDisplaySettings (&g_newMode, CDS_UPDATEREGISTRY | CDS_GLOBAL);
    modeset = ChangeDisplaySettings (&g_newMode, CDS_RESET);

    if (modeset != DISP_CHANGE_SUCCESSFUL)
    {
        // Try again with a lowered refresh rate and increased BPP.  8BPP is 
        // not supported while rotating, and some of the upper display modes can be 
        // filtered by the operating system.
        //
        g_newMode.dmDisplayFrequency = 60;
        if (g_newMode.dmBitsPerPel < 16)
            g_newMode.dmBitsPerPel = 16;

        modeset = ChangeDisplaySettings (&g_newMode, CDS_UPDATEREGISTRY | CDS_GLOBAL);
        modeset = ChangeDisplaySettings (&g_newMode, CDS_RESET);
    }

    if (modeset != DISP_CHANGE_SUCCESSFUL)
    {
        // Put things back the way they were (or try to).  This implies that we 
        // revert the driver registry settings and change the mode registry settings.
        //
        dc = GetDC(NULL);
        retval = ExtEscape (dc, GEODELX_SET_REGISTRY_SETTINGS, sizeof(g_oldSettings), 
            (LPCSTR)&g_oldSettings, sizeof(g_oldSettings), (LPSTR)&g_oldSettings);
        ReleaseDC (NULL, dc);

        mode = 0;
        while (EnumDisplaySettings (NULL, mode, &dm))
            mode++;

        ChangeDisplaySettings (&g_curMode, CDS_UPDATEREGISTRY | CDS_GLOBAL);
        modeset = ChangeDisplaySettings (&g_curMode, CDS_RESET);
        return -3;
    }

    // 0 indicates success
    //
    return 0;
}
